/* 

dvbd.c, control program 0.03.2, (c)2000 Karsten.Siebert@dataplanet.net

May be copied or modified under the terms of the GNU General Public License.
See /usr/src/linux/COPYING for more information.

04.08.2001:	Jan Rafaj	dvb@3Rcom.org

- occurences of strcpy() replaced with strncpy() to avoid overruns.
- removed "eth_netdevice" - now, its parameter is to be specified
  as an argument to "filter_0".
- added help for command-line flags. Displayed when "-h" or "--help" given.
- added new commandline flag "-b" that allows to specify dvb device name.
- added ability to specify the device or ip incl. optional bitmask, to generate
  DVB device's unicast filter from (through "-e" commandline flag or in config
  file, as 2nd parameter of "filter_0", plus a possibility to omit this
  parameter completely - then, use MACaddr of DVB card for filter_0).

04.04.2000:

- added MAC byte validation in /etc/dvbd.conf

18.03.2000:

- added UDP socket reader for bit filter setting via network using mrouted

07.03.2000:

- added software watchdog parameter (-w to enable/default none),
- software watchdog parameter soft_margin has to be larger than cnt (300 sec),

*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include <syslog.h>
#include <ctype.h>
#include <netdb.h>
#include <dirent.h>
#include <netinet/in.h>
#include <linux/if.h>
#include <linux/if_arp.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/param.h>
#include <arpa/inet.h>

// This actually needs to be included from Siemens-DVB drivers.
#include "./../DVB/driver/dvb_v4l.h"

#include "dvbd.h"


int main (int argc, char *argv[])
{
  struct frontend front ;
  struct sockaddr_in locale ;
  struct sockaddr_in remote ;
  fd_set fds ;
  long cnt = 1L ;
  int  sd  = -1 ;
  int  nn  = -1 ;
  int  vid = -1 ;
  int  ret = -1 ;
  int  wdt = -1 ;

  StopFlag  = 0 ;
  TraceFlag = 1 ;

  memset (&dvbcfg, 0, sizeof (dvbcfg)) ;

// ioctl(vid, VIDIOCGFRONTEND, &front) ; 
// readcfg ("", &front) ;

  front.fec=8 ;
  front.diseqc=0 ;
  front.srate=22000000 ;
  front.video_pid=0 ;
  front.audio_pid=0 ;
  front.AFC=0 ;
  front.channel_flags=DVB_CHANNEL_FTA ;
      
  while ((nn = getopt (argc, argv, "ha:b:c:d:e:f:p:qs:t:v:w")) != -1)
  {
    switch (nn)
    {
      case 'h':
	printf("DVBD - daemon for setting up DVB card parameters\n\n") ;
	printf("Usage: %s\t[-a <0|1>] [-b <devname>] [-c <cfgfile>] [-d <0|1>]\n"              , argv[0]) ;
	printf("\t\t[-e <device|IP[/bitmask]>] [-f <freq>] [-p <0|1>] [-q]\n") ;
	printf("\t\t[-s <symrate>] [-t <0|1>] [-v <0|1>] [-w]\n\n") ;
	printf("Where: -a = AFC. 0 = OFF, 1 = ON. Default = OFF.\n") ;
	printf("       -b = name of DVB-device to use. Default = dvb0.\n") ;
	printf("       -c = path to config file.\n") ;
	printf("            Default = /etc/dvbd.conf\n") ;
	printf("       -d = DiSEqC. 0 = OFF, 1 = ON. Default = OFF.\n") ;
	printf("       -e = name of network device, or IP address [with\n") ;
	printf("            optional net bitmask], from which to calculate\n") ;
	printf("            the value of unicast filter_0. Default = none\n") ;
	printf("            => use real MAC address of DVB card.\n") ;
	printf("       -f = receiver frequency in Hz. Integer value between\n");
	printf("            10600000 and 14500000. Default = frequency\n") ;
	printf("            of transponder 115 on ASTRA 1G = 12662750 Hz.\n") ;
	printf("       -p = power to LNB. 0 = OFF, 1 = ON. Default = ON.\n") ;
	printf("       -q = do not run DVBD in debug mode. Default = do.\n") ;
	printf("       -s = symbol rate in CPS. Value between 1500000\n") ;
	printf("            and 46000000. Default = 22000000 CPS.\n") ;
	printf("       -t = 22 kHz signal to LNB. 0 = OFF, 1 = ON.\n") ;
	printf("            Default = ON.\n") ;
	printf("       -v = polarisation. 0 = VERTICAL, 1 = HORIZONTAL.\n") ;
	printf("            Default = HORIZONTAL.\n") ;
	printf("       -w = use /dev/watchdog as time schedule.\n") ;
	printf("            Default = do not use.\n") ;
	printf("After debugging, DVBD should be run with -q parameter.\n\n") ;
	printf("Commandline flags can be used as alternative to dvbd.conf,\n") ;
	printf("but will be replaced with values from dvbd.conf, if these\n") ;
	printf("exist and are valid.\n\n") ;
	printf("Default values are used in case when config file does not\n") ;
	printf("exist or a specific parameter couldnt be found as cmdline\n") ;
	printf("parameter, or when the given parameter is supplied with\n") ;
	printf("bogus value.\n\n") ;
	exit(0) ;

      case 'a':
        ret = atoi (optarg) ;

        if (ret == 0 || ret == 1)
        {
          front.AFC = ret ;
        }
        break ;

      case 'b':
        if (valid_ifname (optarg) == 1)
        {
          strncpy (dvb_ifname, optarg, sizeof (dvb_ifname)) ;
        }
        break ;

      case 'c':
        strncpy (cfgname, optarg, sizeof (cfgname)) ;
        break ;

      case 'd':
        ret = atoi (optarg) ;

        if (ret == 0 || ret == 1)
        {
          front.diseqc = ret ;
        }
        break ;

      case 'e':
        if ((valid_ifname (optarg) == 1) || (valid_bitmask (optarg) == 1))
        {
          strncpy (filter0_ifname_or_ip, optarg, sizeof (filter0_ifname_or_ip));
        }
        break ;

      case 'f':
        ret = atoi (optarg) ;

        if (ret > 10600000 && ret < 14500000)
        {
          front.freq = front.curfreq = ((__u32) ret - 10600000) * 1000 ;
        }
        break ;

      case 'p':
        ret = atoi (optarg) ;

        if (ret == 0 || ret == 1)
        {
          front.power = ret ;
        }
        break ;

      case 'q':
        TraceFlag = 0 ;
        break ;

      case 's':
        ret = atoi (optarg) ;

        if (ret > 1500000 && ret < 46000000)
        {
          front.srate = (__u32) ret ;
        }
        break ;

      case 't':
        ret = atoi (optarg) ;

        if (ret == 0 || ret == 1)
        {
          front.ttk = ret ;
        }
        break ;

      case 'v':
        ret = atoi (optarg) ;

        if (ret == 0 || ret == 1)
        {
          front.volt = ret ;
        }
        break ;

      case 'w':
        wdt = 1 ;
        break ;
    
      default:
        break ;
    }
  }

  if (TraceFlag == 0)
  {
    MakeDaemon (LoadFdsd) ;
  }


  openlog ("dvbd", LOG_PID, LOG_NEWS) ;

  syslog (LOG_NOTICE, "starting") ;

  if (wdt == 1)
  {
    if ((wdt = open ("/dev/watchdog", O_WRONLY)) == -1)
    {
      syslog (LOG_ERR, "watchdog: %m") ;
    }
  }

  if ((vid = open ("/dev/video", O_RDWR)) == -1)
  {
    syslog (LOG_ERR, "open: %m") ;
    fflush (stdout) ;    
    goto finish ;
  }


  ioctl (vid, VIDIOCGFRONTEND, &front);

  if (readcfg (cfgname, &front) < 0)
  {
    syslog (LOG_ERR, "readcfg: error\n") ;
    fflush (stdout) ;    
    goto finish ;
  }


  front.fec = 8 ;
        
  if (ioctl (vid, VIDIOCSFRONTEND, &front) == -1)
  {
    syslog (LOG_ERR, "ioctl VSFRONTEND: %m") ;
     fflush (stdout) ;
    goto finish ;  
  }
   
  memset ((char *) &locale, 0, sizeof (locale)) ;
  memset ((char *) &remote, 0, sizeof (remote)) ;

  locale.sin_family      = AF_INET ;
  locale.sin_port        = htons (1060) ;
  locale.sin_addr.s_addr = htonl (INADDR_ANY) ;

  if ((sd = socket (PF_INET, SOCK_DGRAM, 0)) == -1)
  {
    syslog (LOG_ERR, "socket: %m") ;
    goto finish ;
  }

  if (bind (sd, (struct sockaddr *) &locale, sizeof (locale)) == -1)
  {
    syslog (LOG_ERR, "bind: %m") ;
    goto finish ;
  }

  if (setcfgfilter (vid) == 0)
  {
    while (StopFlag == 0)
    {
      struct timeval tout ;

      FD_ZERO (&fds) ;
      FD_SET (sd, &fds) ;

      tout.tv_sec  = cnt ;
      tout.tv_usec = 0L ;

      if ((nn = select (sd + 1, &fds, (fd_set *) 0, (fd_set *) 0, &tout)) < 0)
      {
        if (errno != EINTR)
        {
          syslog (LOG_ERR, "select: %m") ;
        }
      }

      else if (nn == 0)
      {
        if (ioctl (vid, VIDIOCGFRONTEND, &front) == -1)
        {
          syslog (LOG_ERR, "ioctl VGFRONTEND: %m") ;
	  fflush (stdout) ;
        }

        if (TraceFlag == 1)
        {
          cnt = 1L ;

          printf ("Fsync = %u, Freq = %u, Sync = %u Afc = %u Agc = %u Nest = %u Vber = %u    \r", front.fsync, 
                    front.freq , front.sync, front.afc, 
		    255 - (char)((char)(front.agc>>4) + (int)128), front.nest,  front.vber) ;

          fflush (stdout) ;
        } 

        else 
        {
          cnt = 300L ;

          if (front.sync == 127)
          {
            syslog (LOG_NOTICE, "Sync[%d]: ok", front.sync) ;
          }

          else
          {
            syslog (LOG_NOTICE, "Sync[%d]: failure", front.sync) ;
	    printf("Sync[%d]: failure", front.sync);	    
          }
        }

        if (wdt != -1)
        {
          write (wdt, "\0", 1) ;
        }
      }

      else if (nn > 0)
      { 
        if (FD_ISSET (sd, &fds))
        {
          int  len = sizeof (struct sockaddr_in) ;
          char buff[100] ;

          if ((nn = recvfrom (sd, buff, 100, 0, 
                             (struct sockaddr *) &remote, &len)) < 0) 
          {
            if (errno != EINTR)
            {
              syslog (LOG_ERR, "recvfrom: %m") ;
            }
          }

          else if (nn > 0)
          {
            int n = -1 ;
            struct dvbbuffer *p = (struct dvbbuffer *) buff ;
            struct frontend  *f = NULL ;
         
            switch (p->status)
            {
              case SET_DYNAMO:
              case SET_FILTER:
                if ((n = setfilter (vid, p)) == -1)
                {
                  syslog (LOG_ERR, "setfilter: error") ;
                }
                break ;

              case REL_FILTER:
                if ((n = relfilter (vid, p)) == -1)
                {
                  syslog (LOG_ERR, "relfilter: error") ;
                }
                break ;

              case SET_SATPAR:
                f = (struct frontend *) (buff + sizeof (struct dvbbuffer)) ;

                if (ioctl (vid, VIDIOCSFRONTEND, f) == -1)
                {
                  syslog (LOG_ERR, "ioctl VSFRONTEND: %m") ;
                }
                break ;

              case GET_SATPAR:
                f = (struct frontend *) (buff + sizeof (struct dvbbuffer)) ;

                if (ioctl (vid, VIDIOCGFRONTEND, f) == -1)
                {
                  syslog (LOG_ERR, "ioctl VGFRONTEND: %m") ;
                }
                break ;

              default:
                break;
            }
          }

          else if (nn == 0)
          {
          }
        }
      }
    }
  }

finish : 

  syslog (LOG_NOTICE, "exiting") ;

  if (wdt != -1)
  {
    (void) close (wdt) ;
  }

  if (sd != -1)
  {
    (void) close (sd) ;
  }

  if (vid != -1)
  {
    relcfgfilter (vid) ;
    (void) close (vid) ;
  }

  closelog () ;
  fflush (stdout) ;
  return errno ;
}


int SetSignal (void (*sigquit) (void), void (*sigchld) (void),
               void (*sigintr) (void), void (*sigterm) (void),
               void (*sighaup) (void), void (*sigalrm) (void))
{
  struct sigaction quit_act ;
  struct sigaction chld_act ;
  struct sigaction intr_act ;
  struct sigaction term_act ;
  struct sigaction haup_act ;
  struct sigaction alrm_act ;
  sigset_t         set ;

  if (sigquit != NULL)
  {
    Sigquit = sigquit ;
  }

  if (sigchld != NULL)
  {
    Sigchld = sigchld ;
  }

  if (sigintr != NULL)
  {
    Sigintr = sigintr ;
  }

  if (sigterm != NULL)
  {
    Sigterm = sigterm ;
  }

  if (sighaup != NULL)
  {
    Sighaup = sighaup ;
  }

  if (sigalrm != NULL)
  {
    Sigalrm = sigalrm ;
  }

  sigemptyset (&set) ;

  quit_act.sa_flags   = 0 ;
  quit_act.sa_mask    = set ;
  quit_act.sa_handler = SigQuit ; 

  sigaction (SIGQUIT, &quit_act, (struct sigaction *) 0) ;

  sigemptyset (&set) ;

  chld_act.sa_flags   = 0 ;
  chld_act.sa_mask    = set ;
/*
  chld_act.sa_handler = SigChld ; 
*/
  chld_act.sa_handler = SIG_IGN ; 

  sigaction (SIGCHLD, &chld_act, (struct sigaction *) 0) ;

  sigemptyset (&set) ;

  intr_act.sa_flags   = 0 ;
  intr_act.sa_mask    = set ;
  intr_act.sa_handler = SigIntr ; 

  sigaction (SIGINT, &intr_act, (struct sigaction *) 0) ;

  sigemptyset (&set) ;

  term_act.sa_flags   = 0 ;
  term_act.sa_mask    = set ;
  term_act.sa_handler = SigTerm ; 

  sigaction (SIGTERM, &term_act, (struct sigaction *) 0) ;

  sigemptyset (&set) ;

  haup_act.sa_flags   = 0 ;
  haup_act.sa_mask    = set ;
  haup_act.sa_handler = SigHaup ; 

  sigaction (SIGHUP, &haup_act, (struct sigaction *) 0) ;

  sigemptyset (&set) ;

  alrm_act.sa_flags   = 0 ;
  alrm_act.sa_mask    = set ;
  alrm_act.sa_handler = SigAlrm ; 

  sigaction (SIGALRM, &alrm_act, (struct sigaction *) 0) ;

  return 1 ;
}

void SigQuit ()
{
  Sigquit () ;
}

/*
void SigChld ()
{
  Sigchld () ;
}
*/

void SigIntr ()
{
  Sigintr () ;
}

void SigTerm ()
{
  Sigterm () ;
}

void SigHaup ()
{
  Sighaup () ;
}

void SigAlrm ()
{
  Sigalrm () ;
}

void SigNoop ()
{
}

void ChldProc ()
{
  wait (NULL) ;
}

void StopProc ()
{
  StopFlag = 1 ;
}

void LoadFdsd ()
{
  syslog (LOG_NOTICE, "sighaup arrived") ;
}

void MakeDaemon (void (*lpProc) (void)) 
{
  pid_t childpid ;
  int   fd ;

  if ((childpid = fork ()) < 0)
  {
    syslog (LOG_ERR, "%m") ;
  }

  else if (childpid > 0)
  {
    exit (0) ;
  }

  if (setpgrp () == -1)
  {
    syslog (LOG_ERR, "setpgrp: %m") ;
  }

  SetSignal (NULL, NULL, NULL, NULL, (void*) SIG_IGN, NULL) ;

  if ((childpid = fork ()) < 0)
  {
    syslog (LOG_ERR, "fork: %m") ;
  }

  else if (childpid > 0)
  {
    exit (0) ;
  }

  for (fd = 0; fd < NOFILE; fd++)
  {
    (void) close (fd) ;
  }

  errno = 0 ;

  umask (0) ;

  (void *) open ("/", 2) ;
  (void *) dup2 (0, 1) ;
  (void *) dup2 (0, 2) ;

#ifdef SYSV
  (void) setpgrp () ;
#else
#ifdef TIOCNOTTY
  fd = open ("/dev/tty", 2) ;

  if (fd >= 0)
  {
    (void) ioctl (fd, TIOCNOTTY, (char *) 0) ;
    (void) close (fd) ;
  }
#else
  if (setsid () < 0)
  {
    syslog (LOG_ERR, "setsid: %m") ;
  }
#endif
#endif

  SetSignal (NULL, ChldProc, NULL, StopProc, lpProc, NULL) ;
}


int setfilter (int vd, struct dvbbuffer *p)
{
  int n = 0 ;

  /* Check if the filter already exists */

  for (n = 0; n < MAX_FILTER; n++)
  {
    /* IP multicast address or IP unicast address */

    if ((p->ipaddr>0xe0000000) && (dvbcfg[n].mcaddr==(p->ipaddr & 0x007fffff)))
    {
      dvbcfg[n].number++ ;
      return -2 ;
    }

    else if (dvbcfg[n].mcaddr == p->ipaddr)
    {
      dvbcfg[n].number++ ;
      return -2 ;
    }
  }

  for (n = 0; n < MAX_FILTER; n++)
  {
    /* Free and NOT set through configuration file */

    if (dvbcfg[n].status == OFF)
    {
      /* RFC 1112 ip address part */

      if (p->ipaddr > 0xe0000000)
      {
        dvbcfg[n].mcaddr = (p->ipaddr & 0x007fffff) ;
      }

      else
      {
        dvbcfg[n].mcaddr = p->ipaddr ;
      }

      dvbcfg[n].filter.data[0] = 0x3eff ;

      dvbcfg[n].filter.mode    = 0x0c ;

      /* data planet international recommendation */

      if (p->status == SET_DYNAMO)
      {
        dvbcfg[n].filter.pid     = dvbcfg[n].mcaddr & 0x1fff ;

        dvbcfg[n].filter.data[1] = (dvbcfg[n].mcaddr << 8) | 0x00ff ;
        dvbcfg[n].filter.data[2] = (dvbcfg[n].mcaddr)      | 0x00ff ;
        dvbcfg[n].filter.data[6] = (dvbcfg[n].mcaddr >> 8) | 0x00ff ;

        /* RFC 1112 mac address part */

        dvbcfg[n].filter.data[7] = 0x5eff ;
        dvbcfg[n].filter.data[8] = 0x00ff ;
        dvbcfg[n].filter.data[9] = 0x01ff ;
      }

      else
      {
        dvbcfg[n].filter.pid     = p->usrpid ;

        dvbcfg[n].filter.data[1] = p->mac[5] ;
        dvbcfg[n].filter.data[2] = p->mac[4] ;
        dvbcfg[n].filter.data[6] = p->mac[3] ;
        dvbcfg[n].filter.data[7] = p->mac[2] ;
        dvbcfg[n].filter.data[8] = p->mac[1] ;
        dvbcfg[n].filter.data[9] = p->mac[0] ;
      }

      if (ioctl (vd, VIDIOCSBITFILTER, &dvbcfg[n].filter) < 0)
      {
        syslog (LOG_ERR, "ioctl SETBFILTER[%d]: %m", n) ;
        return -1 ;
      }

      dvbcfg[n].number++ ;

      dvbcfg[n].status = ON ;

      return dvbcfg[n].filter.handle ;
    }
  }

  return -2 ;
}


int relfilter (int vd, struct dvbbuffer *p)
{
  int n ;

  for (n = 0; n < MAX_FILTER; n++)
  {
    /* Filter set through configuration CANNOT be removed */

    if (dvbcfg[n].status == ON)
    {
      if ((dvbcfg[n].mcaddr == p->ipaddr) || 
          (dvbcfg[n].mcaddr == (p->ipaddr & 0x007fffff)))
      {
        if (dvbcfg[n].number == 1)
        {
          if (ioctl (vd, VIDIOCSSHUTDOWNFILTER, &dvbcfg[n].filter.handle) < 0)
          {
            syslog (LOG_ERR, "ioctl SHUTDOWN[%d]: %m", n) ;
            return -1 ;
          }

          dvbcfg[n].status = OFF ;

          memset (&dvbcfg[n], 0, sizeof (struct dvbfilter)) ;

          return n ;
        }
  
        else
        {
          dvbcfg[n].number-- ;
        }
      }
    }
  }

  return -2 ;
}


int setcfgfilter (int vd)
{
  int n ;

  for (n = 0; n < MAX_FILTER; n++)
  {
    if (dvbcfg[n].status == ON)
    {
      if (ioctl (vd, VIDIOCSBITFILTER, &dvbcfg[n].filter) < 0)
      {
        syslog (LOG_ERR, "ioctl SETBFILTER[%d]: %m", n) ;
      }
    }
  }

  return 0 ;
}


int relcfgfilter (int vd)
{
  int n = 0 ;

  for (n = 0; n < MAX_FILTER; n++)
  {
    if (dvbcfg[n].status == ON)
    {
      if (ioctl (vd, VIDIOCSSHUTDOWNFILTER, &dvbcfg[n].filter.handle) < 0)
      {
        syslog (LOG_ERR, "ioctl SHUTDOWN[%d]: %m", n) ;
      }
    
      dvbcfg[n].status = OFF ;
    }
  }

  return 0 ;
}


int resolve_ether (char *buf, unsigned char *addr)
{
  unsigned char *ptr = addr ;
  int i = 0 ;

  while ((*buf != '\0') && (i < ETH_ALEN)) 
  {
    int val = 0 ;
    int c   = *buf++ ;

    if (isdigit (c))
    {
      val = c - '0' ;
    }
    else if (c >= 'a' && c <= 'f') 
    {
      val = c - 'a' + 10 ;
    }
    else if (c >= 'A' && c <= 'F') 
    {
      val = c - 'A' + 10 ;
    }
    else 
    {
      errno = EINVAL ;
      return -1 ;
    }

    val <<= 4 ;
    c = *buf++ ;

    if (isdigit (c)) 
    {
      val |= c - '0' ;
    }
    else if (c >= 'a' && c <= 'f') 
    {
      val |= c - 'a' + 10 ;
    }
    else if (c >= 'A' && c <= 'F') 
    {
      val |= c - 'A' + 10 ;
    }
    else 
    {
      errno = EINVAL ;
      return -1 ;
    }

    *ptr++ = (unsigned char) (val & 0xff) ;
    i++ ;

    if (*buf == ':') 
    {
      buf++ ;
    }
  }

  return 0 ;
}


int ipget (unsigned char *ip, unsigned char *mask, char *dev_or_ip)
{
  char *p = NULL ;
  int s, i ;
  struct ifreq req ;
  struct in_addr ins ;

  mask[0] = mask[1] = mask[2] = mask[3] = 0xff ;

  s = socket (AF_INET, SOCK_DGRAM, 0) ;

  strncpy (req.ifr_name, dev_or_ip, sizeof (req.ifr_name)) ;

  if (ioctl (s, SIOCGIFADDR, &req) < 0)
  {
    close (s) ;
  }

  else
  {
    memcpy (ip, req.ifr_addr.sa_data + sizeof (unsigned short), 4) ;
    return 0 ;
  }

  strtok (dev_or_ip, "/") ;

  if (inet_aton (dev_or_ip, &ins) == 0)
  {
    return -1 ;
  }

  else
  {
    for (i = 0; i < 4; i++)
    {
     ip[i] = (ins.s_addr >> (i * 8)) & 0xff ;
    }

    p = strtok (NULL, "/") ;

    if (p)
    {
      i = atoi (p) ;

      if (i < 0 || i > 32)
      {
        i = 32 ;
      }

      /* convert the bitmask to IP value in network format */
      ins.s_addr = htonl (0xffffffff ^
                          ((0x00000001 << (32 - i)) - (i == 0 ? 2 : 1)) ) ;

      for (i = 0; i < 4; i++)
      {
        mask[i] = (ins.s_addr >> (i * 8)) & 0xff ;
      }
    }

    return 0 ;
  }
}

int setmac (unsigned char *ip)
{
  struct ifreq ifr ;
  unsigned char mac[6] ;
  int s ;
  struct sockaddr sa ;
    
  s = socket (AF_INET, SOCK_DGRAM, 0) ;

  // Interface down

  strncpy (ifr.ifr_name, dvb_ifname, sizeof (ifr.ifr_name)) ;

  if (ioctl (s, SIOCGIFFLAGS, &ifr) < 0)
  {
    syslog (LOG_ERR, "GetFlags: %m") ;
    return -1 ;
  }

  strncpy (ifr.ifr_name, dvb_ifname, sizeof (ifr.ifr_name)) ;

  ifr.ifr_flags &= ~0x01 ;

  if (ioctl (s, SIOCSIFFLAGS, &ifr) < 0) 
  {
    syslog (LOG_ERR, "SetFlags: %m") ;
    return -1 ;
  }

  // set MAC
 
  mac[0] = 0x00 ;
  mac[1] = 0x02 ;
  mac[2] = ip[0] ;
  mac[3] = ip[1] ;
  mac[4] = ip[2] ;
  mac[5] = ip[3] ;

  sa.sa_family = ARPHRD_ETHER ;
  memcpy (sa.sa_data, mac, 6) ;

  strncpy (ifr.ifr_name, dvb_ifname, sizeof (ifr.ifr_name)) ;

  memcpy ((char *) &ifr.ifr_hwaddr, (char *) &sa, sizeof (struct sockaddr)) ;

  if (ioctl (s, SIOCSIFHWADDR, &ifr) < 0) 
  {
    syslog (LOG_ERR, "SetHWAddr: %m") ;
    return -1 ;
  }

  // Interface up again
      
  strncpy (ifr.ifr_name, dvb_ifname, sizeof (ifr.ifr_name)) ;

  if (ioctl (s, SIOCGIFFLAGS, &ifr) < 0) 
  {
    syslog (LOG_ERR, "GetFlags: %m") ;
    return -1 ;
  }

  strncpy (ifr.ifr_name, dvb_ifname, sizeof (ifr.ifr_name)) ;

  ifr.ifr_flags |= 0x01 ;

  if (ioctl (s, SIOCSIFFLAGS, &ifr) < 0) 
  {
    syslog (LOG_ERR, "SetFlags: %m") ;
    return -1 ;
  }

  return 0 ;
}

int getmac_dvbdevice (char *mac)
{
  int s ;
  struct ifreq ifr ;

  s = socket (AF_INET, SOCK_DGRAM, 0) ;

  strncpy (ifr.ifr_name, dvb_ifname, sizeof (ifr.ifr_name)) ;

  if (ioctl (s, SIOCGIFHWADDR, &ifr) < 0)
  {
    syslog (LOG_ERR, "GetMac: %m") ;
    return -1 ;
  }

  memcpy (mac, ifr.ifr_hwaddr.sa_data, 6) ;

  return 0 ;
}

int valid_ifname (char *dev_ptr)
{
  int s ;
  struct ifreq ifr ;

  s = socket (AF_INET, SOCK_DGRAM, 0) ;

  strncpy (ifr.ifr_name, dev_ptr, sizeof (ifr.ifr_name)) ;

  if (ioctl (s, SIOCGIFFLAGS, &ifr) >= 0)
  {
    return 1 ;
  }

  return 0 ;
}

int valid_bitmask (char *bitmask_ptr)
{
  char *p = NULL ;
  int retcode = 0 ;
  struct in_addr ins ;

  p = index (bitmask_ptr, '/') ;

  if (p)
  {
    *p = '\0' ;
  }

  if (inet_aton (bitmask_ptr, &ins) == 1)
  {
    retcode = 1 ;
  }

  if (p)
  {
    *p = '/' ;
  }

  return retcode ;
}

int readcfg (char *cfg, struct frontend *fend)
{
  FILE *fp = NULL ;
  char buff[BUFSIZE + 1] ;
  char *v = NULL ;
  char *s = NULL ;

  fend->type = FRONT_DVBS ;

  if (!strlen (cfg))
  {
    int n ;
    for (n= 0; n < MAX_FILTER; n++)
    {
      dvbcfg[n].status = OFF ;
    }

    fend->ttk    = 1 ;
    fend->AFC    = 0 ;
    fend->diseqc = 0 ;
    fend->volt   = 1 ;
    fend->power  = 1 ;

    /* Default value is TR 115 on ASTRA 1G/H */
    fend->fec=8 ;
    fend->srate  = 22000000 ;
    fend->freq   = fend->curfreq = (12662750 - 10600000) * 1000 ;

    fend->video_pid = 0 ;
    fend->audio_pid = 0 ;

    return 0 ;
  }

  if ((fp = fopen (cfg, "r")) == NULL)
  {
    syslog (LOG_ERR, "No config file (%s)... using program defaults.",
           cfg) ;
    return 0 ;
  }

  while (fgets (buff, BUFSIZE, fp) > 0)
  {
    v = NULL ;
    s = NULL ;

    strtok (buff, "\n") ;

    v = strtok (buff, " ") ;
    s = strtok (NULL, " ") ;

    if (strcmp (v, "dvb_ifname") == 0)
    {
      if (s != NULL)
      {
        if (valid_ifname (s) == 1)
        {
          strncpy (dvb_ifname, s, sizeof (dvb_ifname)) ;
        }
      }
    }

    if (strcmp (v, "ttk") == 0)
    {
      if (s != NULL)
      {
        fend->ttk = atoi (s) ;

        if ((fend->ttk != 0) && (fend->ttk != 1))
        {
          fend->ttk = 1 ;
        }
      }

      else
      {
        fend->ttk = 1 ;
      }
    }

    if (strcmp (v, "AFC") == 0)
    {
      if (s != NULL)
      {
        fend->AFC = atoi (s) ;

        if ((fend->AFC != 0) && (fend->AFC != 1))
        {
          fend->AFC = 1 ;
        }
      }

      else
      {
        fend->AFC = 1 ;
      }
    }

    if (strcmp (v, "diseqc") == 0)
    {
      if (s != NULL)
      {
        fend->diseqc = atoi (s) ;

        if ((fend->diseqc != 0) && (fend->diseqc != 1))
        {
          fend->diseqc = 0 ;
        }
      }

      else
      {
        fend->diseqc = 0 ;
      }
    }

    if (strcmp (v, "polarisation") == 0)
    {
      if (s != NULL)
      {
        fend->volt = atoi (s) ;

        if ((fend->volt != 0) && (fend->volt != 1))
        {
          fend->volt = 1 ;
        }
      }

      else
      {
        fend->volt = 1 ;
      }
    }

    if (strcmp (v, "power") == 0)
    {
      if (s != NULL)
      {
        fend->power = atoi (s) ;

        if ((fend->power != 0) && (fend->power != 1))
        {
          fend->power = 1 ;
        }
      }

      else
      {
        fend->power = 1 ;
      }
    }

    if (strcmp (v, "symbolrate") == 0)
    {
      __u32 ret ;

      if (s != NULL)
      {
        ret = (__u32) atoi (s) ;

        if ((ret > 1500000) && (ret < 46000000))
        {
          fend->srate = ret ;
        }
      }

      else
      {
        fend->srate = 22000000 ;
      }
    }

    if (strcmp (v, "frequency") == 0)
    {
      __u32 ret ;

      if (s != NULL)
      {
        ret = (__u32) atoi (s) ;

        if ((ret > 10600000) && (ret < 14500000))
        {
          fend->freq = fend->curfreq = ((__u32) ret - 10600000) * 1000 ;
        }
      }

      else
      {
        fend->freq = fend->curfreq = (12662750 - 10600000) * 1000 ;
      }
    }

    if (strcmp (v, "filter_0") == 0)
    {
      if (s != NULL)
      {
        char *m = NULL ;

        dvbcfg[0].status = ON ;

        dvbcfg[0].filter.data[0] = 0x3eff ;

        dvbcfg[0].filter.pid     = (__u16) atoi (s) ;

        dvbcfg[0].filter.mode    = 0x0c ;

        m = strtok (NULL, " ") ;

        if (m != NULL)
        {
          unsigned char ip[4] ;
          unsigned char mask[4] ;

          if ((valid_ifname (m) == 1) || (valid_bitmask (m) == 1))
          {
            strncpy (filter0_ifname_or_ip, m, sizeof (filter0_ifname_or_ip)) ;
          }

          if (ipget (ip, mask, filter0_ifname_or_ip) == -1)
          {
            syslog (LOG_NOTICE, "Can't get IP (net) for filter_0. Stop.\n") ;
            return -1 ;
          }

          ip[3] &= mask[3] ;
          ip[2] &= mask[2] ;
          ip[1] &= mask[1] ;
          ip[0] &= mask[0] ;

          syslog(LOG_NOTICE,"IP(net) for filter_0 is %u.%u.%u.%u/%u.%u.%u.%u\n",
                            ip[0], ip[1], ip[2], ip[3],
                            mask[0], mask[1], mask[2], mask[3]) ;

          dvbcfg[0].filter.data[1] = (ip[3] << 8) | mask[3] ;
          dvbcfg[0].filter.data[2] = (ip[2] << 8) | mask[2] ;
          dvbcfg[0].filter.data[6] = (ip[1] << 8) | mask[1] ;
          dvbcfg[0].filter.data[7] = (ip[0] << 8) | mask[0] ;
          dvbcfg[0].filter.data[8] = (0x02  << 8) | 0x00ff ;
          dvbcfg[0].filter.data[9] = (0x00  << 8) | 0x00ff ;

          setmac (ip) ;
        }
        else
        {
          unsigned char mac[6] ;

          if (getmac_dvbdevice (mac) == -1)
          {
            syslog (LOG_NOTICE, "Can't get MAC for filter_0 from %s. Stop.",
                   dvb_ifname) ;
            return -1 ;
          }

          syslog (LOG_NOTICE, "Using MAC address of DVB card for filter_0.\n") ;

          dvbcfg[0].filter.data[1] = (mac[5] << 8) | 0x00ff ;
          dvbcfg[0].filter.data[2] = (mac[4] << 8) | 0x00ff ;
          dvbcfg[0].filter.data[6] = (mac[3] << 8) | 0x00ff ;
          dvbcfg[0].filter.data[7] = (mac[2] << 8) | 0x00ff ;
          dvbcfg[0].filter.data[8] = (mac[1] << 8) | 0x00ff ;
          dvbcfg[0].filter.data[9] = (mac[0] << 8) | 0x00ff ;
        }
      }

      else
      {
        dvbcfg[1].status = OFF ;
      }
    }

    if (strcmp (v, "filter_1") == 0)
    {
      if (s != NULL)
      {
        char *m = NULL ;

        dvbcfg[1].status = ON ;

        dvbcfg[1].mcaddr = htonl (INADDR_ANY) ;

        dvbcfg[1].filter.data[0] = 0x3eff ;

        dvbcfg[1].filter.pid     = (__u16) atoi (s) ;

        dvbcfg[1].filter.mode    = 0x0c ;

        m = strtok (NULL, " ") ;

        if (m != NULL)
        {
          unsigned int  ft = 0 ;
          unsigned char mac[6] ;

          memset (mac, 0, sizeof (mac)) ;

          resolve_ether (m, mac) ;

          m = strtok (NULL, " ") ;
          if (m != NULL)
          {
            ft = atoi (m) ;
          }
          else
          {
            ft = 0xffffffff ;
          }

          if (ft & BIT0)
          {
            dvbcfg[1].filter.data[9] = (mac[0] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[1].filter.data[9] = 0x0000 ;
          }

          if (ft & BIT1)
          {
            dvbcfg[1].filter.data[8] = (mac[1] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[1].filter.data[8] = 0x0000 ;
          }

          if (ft & BIT2)
          {
            dvbcfg[1].filter.data[7] = (mac[2] << 8) | 0x00ff ;
          }

          else
          {       
            dvbcfg[1].filter.data[7] = 0x0000 ;
          }
 
          if (ft & BIT3)
          {
            dvbcfg[1].filter.data[6] = (mac[3] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[1].filter.data[6] = 0x0000 ;
          }

          if (ft & BIT4)
          {
            dvbcfg[1].filter.data[2] = (mac[4] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[1].filter.data[2] = 0x0000 ;
          }

          if (ft & BIT5)
          {
            dvbcfg[1].filter.data[1] = (mac[5] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[1].filter.data[1] = 0x0000 ;
          }
        }

        else
        {
          dvbcfg[1].filter.data[1] = 0x0000 ;
          dvbcfg[1].filter.data[2] = 0x0000 ;
          dvbcfg[1].filter.data[6] = 0x0000 ;
          dvbcfg[1].filter.data[7] = 0x0000 ;
          dvbcfg[1].filter.data[8] = 0x0000 ;
          dvbcfg[1].filter.data[9] = 0x0000 ;
        }
      }

      else
      {
        dvbcfg[1].status = OFF ;
      }
    }

    if (strcmp (v, "filter_2") == 0)
    {
      if (s != NULL)
      {
        char *m = NULL ;

        dvbcfg[2].status = ON ;

        dvbcfg[2].mcaddr = htonl (INADDR_ANY) ;

        dvbcfg[2].filter.data[0] = 0x3eff ;

        dvbcfg[2].filter.pid     = (__u16) atoi (s) ;

        dvbcfg[2].filter.mode    = 0x0c ;

        m = strtok (NULL, " ") ;

        if (m != NULL)
        {
          unsigned int  ft = 0 ;
          unsigned char mac[6] ;

          memset (mac, 0, sizeof (mac)) ;

          resolve_ether (m, mac) ;

          m = strtok (NULL, " ") ;

          if (m != NULL)
          {
            ft = atoi (m) ;
          }
          else
          {
            ft = 0xffffffff ;
          }

          if (ft & BIT0)
          {
            dvbcfg[2].filter.data[9] = (mac[0] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[2].filter.data[9] = 0x0000 ;
          }

          if (ft & BIT1)
          {
            dvbcfg[2].filter.data[8] = (mac[1] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[2].filter.data[8] = 0x0000 ;
          }

          if (ft & BIT2)
          {
            dvbcfg[2].filter.data[7] = (mac[2] << 8) | 0x00ff ;
          }

          else
          {       
            dvbcfg[2].filter.data[7] = 0x0000 ;
          }
 
          if (ft & BIT3)
          {
            dvbcfg[2].filter.data[6] = (mac[3] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[2].filter.data[6] = 0x0000 ;
          }

          if (ft & BIT4)
          {
            dvbcfg[2].filter.data[2] = (mac[4] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[2].filter.data[2] = 0x0000 ;
          }

          if (ft & BIT5)
          {
            dvbcfg[2].filter.data[1] = (mac[5] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[2].filter.data[1] = 0x0000 ;
          }
        }

        else
        {
          dvbcfg[2].filter.data[1] = 0x0000 ;
          dvbcfg[2].filter.data[2] = 0x0000 ;
          dvbcfg[2].filter.data[6] = 0x0000 ;
          dvbcfg[2].filter.data[7] = 0x0000 ;
          dvbcfg[2].filter.data[8] = 0x0000 ;
          dvbcfg[2].filter.data[9] = 0x0000 ;
        }
      }

      else
      {
        dvbcfg[2].status = OFF ;
      }
    }

    if (strcmp (v, "filter_3") == 0)
    {
      if (s != NULL)
      {
        char *m = NULL ;

        dvbcfg[3].status = ON ;

        dvbcfg[3].mcaddr = htonl (INADDR_ANY) ;

        dvbcfg[3].filter.data[0] = 0x3eff ;

        dvbcfg[3].filter.pid     = (__u16) atoi (s) ;

        dvbcfg[3].filter.mode    = 0x0c ;

        m = strtok (NULL, " ") ;

        if (m != NULL)
        {
          unsigned int  ft = 0 ;
          unsigned char mac[6] ;

          memset (mac, 0, sizeof (mac)) ;

          resolve_ether (m, mac) ;

          m = strtok (NULL, " ") ;

          if (m != NULL)
          {
            ft = atoi (m) ;
          }
          else
          {
            ft = 0xffffffff ;
          }

          if (ft & BIT0)
          {
            dvbcfg[3].filter.data[9] = (mac[0] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[3].filter.data[9] = 0x0000 ;
          }

          if (ft & BIT1)
          {
            dvbcfg[3].filter.data[8] = (mac[1] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[3].filter.data[8] = 0x0000 ;
          }

          if (ft & BIT2)
          {
            dvbcfg[3].filter.data[7] = (mac[2] << 8) | 0x00ff ;
          }

          else
          {       
            dvbcfg[3].filter.data[7] = 0x0000 ;
          }
 
          if (ft & BIT3)
          {
            dvbcfg[3].filter.data[6] = (mac[3] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[3].filter.data[6] = 0x0000 ;
          }

          if (ft & BIT4)
          {
            dvbcfg[3].filter.data[2] = (mac[4] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[3].filter.data[2] = 0x0000 ;
          }

          if (ft & BIT5)
          {
            dvbcfg[3].filter.data[1] = (mac[5] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[3].filter.data[1] = 0x0000 ;
          }
        }

        else
        {
          dvbcfg[3].filter.data[1] = 0x0000 ;
          dvbcfg[3].filter.data[2] = 0x0000 ;
          dvbcfg[3].filter.data[6] = 0x0000 ;
          dvbcfg[3].filter.data[7] = 0x0000 ;
          dvbcfg[3].filter.data[8] = 0x0000 ;
          dvbcfg[3].filter.data[9] = 0x0000 ;
        }
      }

      else
      {
        dvbcfg[3].status = OFF ;
      }
    }

    if (strcmp (v, "filter_4") == 0)
    {
      if (s != NULL)
      {
        char *m = NULL ;

        dvbcfg[4].status = ON ;

        dvbcfg[4].mcaddr = htonl (INADDR_ANY) ;

        dvbcfg[4].filter.data[0] = 0x3eff ;

        dvbcfg[4].filter.pid     = (__u16) atoi (s) ;

        dvbcfg[4].filter.mode    = 0x0c ;

        m = strtok (NULL, " ") ;

        if (m != NULL)
        {
          unsigned int  ft = 0 ;
          unsigned char mac[6] ;

          memset (mac, 0, sizeof (mac)) ;

          resolve_ether (m, mac) ;

          m = strtok (NULL, " ") ;

          if (m != NULL)
          {
            ft = atoi (m) ;
          }
          else
          {
            ft = 0xffffffff ;
          }

          if (ft & BIT0)
          {
            dvbcfg[4].filter.data[9] = (mac[0] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[4].filter.data[9] = 0x0000 ;
          }

          if (ft & BIT1)
          {
            dvbcfg[4].filter.data[8] = (mac[1] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[4].filter.data[8] = 0x0000 ;
          }

          if (ft & BIT2)
          {
            dvbcfg[4].filter.data[7] = (mac[2] << 8) | 0x00ff ;
          }

          else
          {       
            dvbcfg[4].filter.data[7] = 0x0000 ;
          }
 
          if (ft & BIT3)
          {
            dvbcfg[4].filter.data[6] = (mac[3] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[4].filter.data[6] = 0x0000 ;
          }

          if (ft & BIT4)
          {
            dvbcfg[4].filter.data[2] = (mac[4] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[4].filter.data[2] = 0x0000 ;
          }

          if (ft & BIT5)
          {
            dvbcfg[4].filter.data[1] = (mac[5] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[4].filter.data[1] = 0x0000 ;
          }
        }

        else
        {
          dvbcfg[4].filter.data[1] = 0x0000 ;
          dvbcfg[4].filter.data[2] = 0x0000 ;
          dvbcfg[4].filter.data[6] = 0x0000 ;
          dvbcfg[4].filter.data[7] = 0x0000 ;
          dvbcfg[4].filter.data[8] = 0x0000 ;
          dvbcfg[4].filter.data[9] = 0x0000 ;
        }
      }

      else
      {
        dvbcfg[4].status = OFF ;
      }
    }

    if (strcmp (v, "filter_5") == 0)
    {
      if (s != NULL)
      {
        char *m = NULL ;

        dvbcfg[5].status = ON ;

        dvbcfg[5].mcaddr = htonl (INADDR_ANY) ;

        dvbcfg[5].filter.data[0] = 0x3eff ;

        dvbcfg[5].filter.pid     = (__u16) atoi (s) ;

        dvbcfg[5].filter.mode    = 0x0c ;

        m = strtok (NULL, " ") ;

        if (m != NULL)
        {
          unsigned int  ft = 0 ;
          unsigned char mac[6] ;

          memset (mac, 0, sizeof (mac)) ;

          resolve_ether (m, mac) ;

          m = strtok (NULL, " ") ;

          if (m != NULL)
          {
            ft = atoi (m) ;
          }
          else
          {
            ft = 0xffffffff ;
          }

          if (ft & BIT0)
          {
            dvbcfg[5].filter.data[9] = (mac[0] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[5].filter.data[9] = 0x0000 ;
          }

          if (ft & BIT1)
          {
            dvbcfg[5].filter.data[8] = (mac[1] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[5].filter.data[8] = 0x0000 ;
          }

          if (ft & BIT2)
          {
            dvbcfg[5].filter.data[7] = (mac[2] << 8) | 0x00ff ;
          }

          else
          {       
            dvbcfg[5].filter.data[7] = 0x0000 ;
          }
 
          if (ft & BIT3)
          {
            dvbcfg[5].filter.data[6] = (mac[3] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[5].filter.data[6] = 0x0000 ;
          }

          if (ft & BIT4)
          {
            dvbcfg[5].filter.data[2] = (mac[4] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[5].filter.data[2] = 0x0000 ;
          }

          if (ft & BIT5)
          {
            dvbcfg[5].filter.data[1] = (mac[5] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[5].filter.data[1] = 0x0000 ;
          }
        }

        else
        {
          dvbcfg[5].filter.data[1] = 0x0000 ;
          dvbcfg[5].filter.data[2] = 0x0000 ;
          dvbcfg[5].filter.data[6] = 0x0000 ;
          dvbcfg[5].filter.data[7] = 0x0000 ;
          dvbcfg[5].filter.data[8] = 0x0000 ;
          dvbcfg[5].filter.data[9] = 0x0000 ;
        }
      }

      else
      {
        dvbcfg[5].status = OFF ;
      }
    }

    if (strcmp (v, "filter_6") == 0)
    {
      if (s != NULL)
      {
        char *m = NULL ;

        dvbcfg[6].status = ON ;

        dvbcfg[6].mcaddr = htonl (INADDR_ANY) ;

        dvbcfg[6].filter.data[0] = 0x3eff ;

        dvbcfg[6].filter.pid     = (__u16) atoi (s) ;

        dvbcfg[6].filter.mode    = 0x0c ;

        m = strtok (NULL, " ") ;

        if (m != NULL)
        {
          unsigned int  ft = 0 ;
          unsigned char mac[6] ;

          memset (mac, 0, sizeof (mac)) ;

          resolve_ether (m, mac) ;

          m = strtok (NULL, " ") ;

          if (m != NULL)
          {
            ft = atoi (m) ;
          }
          else
          {
            ft = 0xffffffff ;
          }

          if (ft & BIT0)
          {
            dvbcfg[6].filter.data[9] = (mac[0] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[6].filter.data[9] = 0x0000 ;
          }

          if (ft & BIT1)
          {
            dvbcfg[6].filter.data[8] = (mac[1] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[6].filter.data[8] = 0x0000 ;
          }

          if (ft & BIT2)
          {
            dvbcfg[6].filter.data[7] = (mac[2] << 8) | 0x00ff ;
          }

          else
          {       
            dvbcfg[6].filter.data[7] = 0x0000 ;
          }
 
          if (ft & BIT3)
          {
            dvbcfg[6].filter.data[6] = (mac[3] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[6].filter.data[6] = 0x0000 ;
          }

          if (ft & BIT4)
          {
            dvbcfg[6].filter.data[2] = (mac[4] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[6].filter.data[2] = 0x0000 ;
          }

          if (ft & BIT5)
          {
            dvbcfg[6].filter.data[1] = (mac[5] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[6].filter.data[1] = 0x0000 ;
          }
        }

        else
        {
          dvbcfg[6].filter.data[1] = 0x0000 ;
          dvbcfg[6].filter.data[2] = 0x0000 ;
          dvbcfg[6].filter.data[6] = 0x0000 ;
          dvbcfg[6].filter.data[7] = 0x0000 ;
          dvbcfg[6].filter.data[8] = 0x0000 ;
          dvbcfg[6].filter.data[9] = 0x0000 ;
        }
      }

      else
      {
        dvbcfg[6].status = OFF ;
      }
    }

    if (strcmp (v, "filter_7") == 0)
    {
      if (s != NULL)
      {
        char *m = NULL ;

        dvbcfg[7].status = ON ;

        dvbcfg[7].mcaddr = htonl (INADDR_ANY) ;

        dvbcfg[7].filter.data[0] = 0x3eff ;

        dvbcfg[7].filter.pid     = (__u16) atoi (s) ;

        dvbcfg[7].filter.mode    = 0x0c ;

        m = strtok (NULL, " ") ;

        if (m != NULL)
        {
          unsigned int  ft = 0 ;
          unsigned char mac[6] ;

          memset (mac, 0, sizeof (mac)) ;

          resolve_ether (m, mac) ;

          m = strtok (NULL, " ") ;

          if (m != NULL)
          {
            ft = atoi (m) ;
          }
          else
          {
            ft = 0xffffffff ;
          }

          if (ft & BIT0)
          {
            dvbcfg[7].filter.data[9] = (mac[0] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[7].filter.data[9] = 0x0000 ;
          }

          if (ft & BIT1)
          {
            dvbcfg[7].filter.data[8] = (mac[1] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[7].filter.data[8] = 0x0000 ;
          }

          if (ft & BIT2)
          {
            dvbcfg[7].filter.data[7] = (mac[2] << 8) | 0x00ff ;
          }

          else
          {       
            dvbcfg[7].filter.data[7] = 0x0000 ;
          }
 
          if (ft & BIT3)
          {
            dvbcfg[7].filter.data[6] = (mac[3] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[7].filter.data[6] = 0x0000 ;
          }

          if (ft & BIT4)
          {
            dvbcfg[7].filter.data[2] = (mac[4] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[7].filter.data[2] = 0x0000 ;
          }

          if (ft & BIT5)
          {
            dvbcfg[7].filter.data[1] = (mac[5] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[7].filter.data[1] = 0x0000 ;
          }
        }

        else
        {
          dvbcfg[7].filter.data[1] = 0x0000 ;
          dvbcfg[7].filter.data[2] = 0x0000 ;
          dvbcfg[7].filter.data[6] = 0x0000 ;
          dvbcfg[7].filter.data[7] = 0x0000 ;
          dvbcfg[7].filter.data[8] = 0x0000 ;
          dvbcfg[7].filter.data[9] = 0x0000 ;
        }
      }

      else
      {
        dvbcfg[7].status = OFF ;
      }
    }

    if (strcmp (v, "filter_8") == 0)
    {
      if (s != NULL)
      {
        char *m = NULL ;

        dvbcfg[8].status = ON ;

        dvbcfg[8].mcaddr = htonl (INADDR_ANY) ;

        dvbcfg[8].filter.data[0] = 0x3eff ;

        dvbcfg[8].filter.pid     = (__u16) atoi (s) ;

        dvbcfg[8].filter.mode    = 0x0c ;

        m = strtok (NULL, " ") ;

        if (m != NULL)
        {
          unsigned int  ft = 0 ;
          unsigned char mac[6] ;

          memset (mac, 0, sizeof (mac)) ;

          resolve_ether (m, mac) ;

          m = strtok (NULL, " ") ;

          if (m != NULL)
          {
            ft = atoi (m) ;
          }
          else
          {
            ft = 0xffffffff ;
          }

          if (ft & BIT0)
          {
            dvbcfg[8].filter.data[9] = (mac[0] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[8].filter.data[9] = 0x0000 ;
          }

          if (ft & BIT1)
          {
            dvbcfg[8].filter.data[8] = (mac[1] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[8].filter.data[8] = 0x0000 ;
          }

          if (ft & BIT2)
          {
            dvbcfg[8].filter.data[7] = (mac[2] << 8) | 0x00ff ;
          }

          else
          {       
            dvbcfg[8].filter.data[7] = 0x0000 ;
          }
 
          if (ft & BIT3)
          {
            dvbcfg[8].filter.data[6] = (mac[3] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[8].filter.data[6] = 0x0000 ;
          }

          if (ft & BIT4)
          {
            dvbcfg[8].filter.data[2] = (mac[4] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[8].filter.data[2] = 0x0000 ;
          }

          if (ft & BIT5)
          {
            dvbcfg[8].filter.data[1] = (mac[5] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[8].filter.data[1] = 0x0000 ;
          }
        }

        else
        {
          dvbcfg[8].filter.data[1] = 0x0000 ;
          dvbcfg[8].filter.data[2] = 0x0000 ;
          dvbcfg[8].filter.data[6] = 0x0000 ;
          dvbcfg[8].filter.data[7] = 0x0000 ;
          dvbcfg[8].filter.data[8] = 0x0000 ;
          dvbcfg[8].filter.data[9] = 0x0000 ;
        }
      }

      else
      {
        dvbcfg[8].status = OFF ;
      }
    }

    if (strcmp (v, "filter_9") == 0)
    {
      if (s != NULL)
      {
        char *m = NULL ;

        dvbcfg[9].status = ON ;

        dvbcfg[9].mcaddr = htonl (INADDR_ANY) ;

        dvbcfg[9].filter.data[0] = 0x3eff ;

        dvbcfg[9].filter.pid     = (__u16) atoi (s) ;

        dvbcfg[9].filter.mode    = 0x0c ;

        m = strtok (NULL, " ") ;

        if (m != NULL)
        {
          unsigned int  ft = 0 ;
          unsigned char mac[6] ;

          memset (mac, 0, sizeof (mac)) ;

          resolve_ether (m, mac) ;

          m = strtok (NULL, " ") ;

          if (m != NULL)
          {
            ft = atoi (m) ;
          }
          else
          {
            ft = 0xffffffff ;
          }

          if (ft & BIT0)
          {
            dvbcfg[9].filter.data[9] = (mac[0] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[9].filter.data[9] = 0x0000 ;
          }

          if (ft & BIT1)
          {
            dvbcfg[9].filter.data[8] = (mac[1] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[9].filter.data[8] = 0x0000 ;
          }

          if (ft & BIT2)
          {
            dvbcfg[9].filter.data[7] = (mac[2] << 8) | 0x00ff ;
          }

          else
          {       
            dvbcfg[9].filter.data[7] = 0x0000 ;
          }
 
          if (ft & BIT3)
          {
            dvbcfg[9].filter.data[6] = (mac[3] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[9].filter.data[6] = 0x0000 ;
          }

          if (ft & BIT4)
          {
            dvbcfg[9].filter.data[2] = (mac[4] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[9].filter.data[2] = 0x0000 ;
          }

          if (ft & BIT5)
          {
            dvbcfg[9].filter.data[1] = (mac[5] << 8) | 0x00ff ;
          }
          else
          {
            dvbcfg[9].filter.data[1] = 0x0000 ;
          }
        }

        else
        {
          dvbcfg[9].filter.data[1] = 0x0000 ;
          dvbcfg[9].filter.data[2] = 0x0000 ;
          dvbcfg[9].filter.data[6] = 0x0000 ;
          dvbcfg[9].filter.data[7] = 0x0000 ;
          dvbcfg[9].filter.data[8] = 0x0000 ;
          dvbcfg[9].filter.data[9] = 0x0000 ;
        }
      }

      else
      {
        dvbcfg[9].status = OFF ;
      }
    }
  }

//jump4 :

  if (fp != NULL)
  {
    fclose (fp) ;
  }

  return 1 ;
}

